%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(erl_mk_internal).
-copyright('Copyright (c) 1991-98 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('arndt@erix.ericsson.se').

%% Parts of this were taken from mk_bifs.erl.

%% The purpose of this file is to generate the data for the functions
%% bif/2 and builtins/0 in erl_internal.erl. The information is taken
%% from erl_internal.src (copied as the first part of erl_internal.erl),
%% erlang.erl (information about Erlang-implemented BIFs) and bif.tab
%% (information about C-implemented BIFs).
%% bif/2 calls the generated function bif_type/2, which may be useful for
%% future use, but currently remains unexported.
%% bif_type(Name, Arity) returns 'true', 'false' or 'undefined' according to
%% whether Name/Arity is an auto-imported BIF, a non-auto-imported BIF,
%% or not a BIF.

-export([generate/1]).

to_string(X) when atom(X) ->
    atom_to_list(X);
to_string(X) when list(X) ->
    X.

%%% generate/1 is meant to be called with
%%% "erl -s erl_mk_internal generate ...".
generate(Args0) ->
    Args = lists:map(fun to_string/1, Args0),
    [Tab, Src, Erlang, Output] = Args,
    copy_file(Src, Output),
    generate(Tab, Src, Erlang, Output),
    halt().

generate(Tab, Src, Erlang, Output) ->
    List = process(Tab),
    Erlang_list = read_autoimport(Erlang),
    All = sorted_list(List ++ Erlang_list),
    {ok, Out} = file:open(Output, [read, write]),
    {ok, _} = file:position(Out, {eof, 0}),
    make(Out, All, filename:basename(Output)).

sorted_list(List) ->
%%  %% We don't do this the right way now, since lists:keysort/2 isn't stable
%%  %% in R3.
%%  lists:keysort(3, lists:sort(List)).
    True = lists:filter(fun({_,_,Auto}) -> Auto end, List),
    False = lists:filter(fun({_,_,Auto}) -> not Auto end, List),
    lists:sort(False) ++ lists:sort(True).

copy_file(From, To) ->
    {ok, Binary} = file:read_file(From),
    ok = file:write_file(To, Binary).

read_autoimport(File) ->
    {ok, Epp} = epp:open(File, []),
    L = read_autoimport(Epp, []),
    epp:close(Epp),
    L.

read_autoimport(Epp, L) ->
    F = epp:parse_erl_form(Epp),
    case F of
	{eof, _Line} ->
	    L;
	{ok, {attribute, _Line, autoimport, List}} ->
	    read_autoimport(Epp, List ++ L);
	{ok, _} ->
	    read_autoimport(Epp, L)
    end.

%% Processes bif.tab.

process(File) ->
    {ok, Fd} = file:open(File, read),
    List = process(Fd, []),
    file:close(Fd),
    List.

process(Fd, Result) ->
    case scan(io:get_line(Fd, none), [], []) of
	eof ->
	    case Result of
		error ->
		    error;
		List ->
		    lists:reverse(Result)
	    end;
	[] ->
	    process(Fd, Result);
	List ->
	    case parse(List) of
		{ok, Tuple} ->
		    process(Fd, [Tuple|Result]);
		error ->
		    process(Fd, error)
	    end
    end.

%% Splits input line into fields.

scan(eof, Token, Result) ->
    eof;
scan([$#|_], Token, Result) ->
    [];
scan([C|Rest], [], Result) when C >= 0, C =< $  ->
    scan(Rest, [], Result);
scan([C|Rest], Token, Result) when C >= 0, C =< $  ->
    scan(Rest, [], [lists:reverse(Token)|Result]);
scan([C|Rest], Token, Result) ->
    scan(Rest, [C|Token], Result);
scan([], [C|Rest], Result) ->
    scan([], [], [[C|Rest]|Result]);
scan([], [], Result) ->
    lists:reverse(Result).


%% Parses field list.

parse([Func, Arity, AutoImport]) ->
    parse(list_to_atom(Func), catch list_to_integer(Arity),
	  list_to_atom(AutoImport));
parse(_) ->
    io:format("Illegal number of fields.~n").

parse(Func, {'EXIT', _}, AutoImport) ->
    io:format("Bad arity.~n"),
    error;
parse(Func, Arity, true) ->
    {ok, {Func, Arity, true}};
parse(Func, Arity, false) ->
    {ok, {Func, Arity, false}};
parse(Func, Arity, Other) ->
    io:format("Bad auto-import field: ~p.~n", [Other]),
    error.

%% Makes output file.

make(Fd, List, Name) ->
    make_bif_type_2(Fd, List),
    make_builtins_0(Fd, List).

make_bif_type_2(Fd, List) ->
    make1(Fd, List),
    io:format(Fd, "bif_type(_, _) -> undefined.~n", []).

make1(Fd, [{Func, Arity, AutoImport}|Rest]) ->
    io:format(Fd, "bif_type(~p, ~p) -> ~p;~n",
	      [Func, Arity, AutoImport]),
    make1(Fd, Rest);
make1(_, []) ->
    ok.

print_builtin(Fd, {Func, Arity, AutoImport}, Comma) ->
    io:format(Fd, "~s~n   ~p", [Comma, {Func, Arity}]).

make_builtins_0(Fd, List) ->
    io:format(Fd, "~n~nbuiltins() -> [", []),
    write_lists_comma(Fd,
		      lists:filter(fun({_,_,Auto}) -> Auto end, List),
		      fun print_builtin/3),
    io:format(Fd, "].~n", []).

%%% Function for printing a comma-separated list of elements.
%%% Calls Fun(Fd, Element, Comma) for each element in the List,
%%% with Comma being "" for the first element, and "," for the others.
write_lists_comma(Fd, List, Fun) ->
    lists:foldl(fun(E, Comma) -> Fun(Fd, E, Comma),
				 ","
		end,
		"",
		List).
